﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Xml.Linq;
using DotNetNuke.Entities.Portals;
using ToSic.Eav;

namespace ToSic.SexyContent.ImportExport
{
    public class XmlImport
    {
        private SexyContent Sexy = new SexyContent(false);
        private int importChangeId;
        private List<XElement> sourceDimensions;
        private string sourceDefaultLanguage;
        private string sourceDefaultDimensionId;


        public bool ImportXml(string xml, List<ExportImportMessage> messages)
        {
            // Parse XDocument from string
            XDocument doc = XDocument.Parse(xml);

            #region Return if not able to process Import

            // Return if no Root Node "SexyContent"
            if (!doc.Elements("SexyContent").Any())
            {
                messages.Add(new ExportImportMessage("The XML file you specified has an incompatible File Version.", ExportImportMessage.MessageTypes.Error));
                return false;
            }

            // Return if Version does not match
            if (new Version(doc.Element("SexyContent").Attribute("MinimumRequiredVersion").Value) > new Version(ImportExport.FileVersion))
            {
                messages.Add(new ExportImportMessage("The XML file you specified has an incompatible File Version.", ExportImportMessage.MessageTypes.Error));
                return false;
            }

            #endregion

            // Create root node "SexyContent"
            XElement xmlSource = doc.Element("SexyContent");

            // All Changes while importing will refer to the same ChangeLog created here
            importChangeId = Sexy.ContentContext.GetChangeLogId(HttpContext.Current.User.Identity.Name);

            // Prepare dimensions
            sourceDimensions = xmlSource.Element("Header").Element("Dimensions").Elements("Dimension").ToList();
            sourceDefaultLanguage = xmlSource.Element("Header").Element("Language").Attribute("Default").Value;
            sourceDefaultDimensionId = sourceDimensions.FirstOrDefault(p => p.Attribute("ExternalKey").Value == sourceDefaultLanguage).Attribute("DimensionID").Value;

            // Import AttributeSets
            if (xmlSource.Elements("AttributeSets").Any())
                ImportXmlAttributeSets(xmlSource, messages);

            // Import Entities
            if (xmlSource.Elements("Entities").Any())
            {
                // Import content items
                foreach (var entity in xmlSource.Elements("Entities").Elements("Entity"))
                    ImportEntity(entity, messages, Sexy.SexyContentAssignmentObjectTypeID);

                Sexy.ContentContext.SaveChanges();
            }

            if (xmlSource.Elements("Templates").Any())
                ImportXmlTemplates(xmlSource, messages);

            return true;
        }

        #region Import AttributeSets

        private void ImportXmlAttributeSets(XElement root, List<ExportImportMessage> messages)
        {
            var attributeSets = root.Element("AttributeSets");

            // Loop through AttributeSets
            foreach (var attributeSet in attributeSets.Elements("AttributeSet"))
            {
                var name = attributeSet.Attribute("Name").Value;
                var attributeSetStaticName = attributeSet.Attribute("StaticName").Value;
                var description = attributeSet.Attribute("Description").Value;


                var existingAttributeSet = Sexy.ContentContext.GetAttributeSet(attributeSetStaticName);
                if (existingAttributeSet != null && !existingAttributeSet.ChangeLogIDDeleted.HasValue)
                {
                    messages.Add(
                        new ExportImportMessage("Content Type '" + name + "' not imported: The same ContentType already exists.", ExportImportMessage.MessageTypes.Warning));
                    continue;
                }

                // Add AttributeSet
                AttributeSet newSet = Sexy.ContentContext.AddAttributeSet(name, description, attributeSetStaticName, SexyContent.AttributeSetScope);

                foreach (XElement xElementAttribute in attributeSet.Element("Attributes").Elements("Attribute"))
                {
                    var staticName = xElementAttribute.Attribute("StaticName").Value;
                    var type = xElementAttribute.Attribute("Type").Value;
                    var isTitle = Boolean.Parse(xElementAttribute.Attribute("IsTitle").Value);

                    // Add Attribute to AttributeSet
                    var newAtt = Sexy.ContentContext.AppendAttribute(newSet.AttributeSetID, staticName,  type, isTitle);

                    // Import Attribute MetaData
                    foreach (var entity in xElementAttribute.Elements("Entity"))
                        ImportEntity(entity, messages, DataSource.AssignmentObjectTypeIdFieldProperties, newAtt.AttributeID);
                }

                messages.Add(new ExportImportMessage("Content Type " + newSet.Name + " successfully imported.", ExportImportMessage.MessageTypes.Information));
            }
        }

        #endregion

        #region Import Templates

        private void ImportXmlTemplates(XElement Root, List<ExportImportMessage> Messages)
        {
            var templates = Root.Element("Templates");

            foreach (var template in templates.Elements("Template"))
            {
                string AttributeSetStaticName = template.Attribute("AttributeSetStaticName").Value;
                AttributeSet Set = Sexy.ContentContext.GetAttributeSet(AttributeSetStaticName);

                Template t = Sexy.TemplateContext.GetNewTemplate();
                t.Name = template.Attribute("Name").Value;
                t.Path = template.Attribute("Path").Value;

                if (Set == null)
                    Messages.Add(
                        new ExportImportMessage("Content Type for Template '" + t.Name + "' could not be found.",
                                                ExportImportMessage.MessageTypes.Warning));
                else
                    t.AttributeSetID = Sexy.ContentContext.GetAttributeSet(AttributeSetStaticName).AttributeSetID;

                string DemoEntityGuid = template.Attribute("DemoEntityGUID").Value;
                if (!String.IsNullOrEmpty(DemoEntityGuid))
                {
                    var EntityGuid = Guid.Parse(DemoEntityGuid);
                    if (Sexy.ContentContext.EntityExists(EntityGuid))
                        t.DemoEntityID = Sexy.ContentContext.GetEntity(EntityGuid).EntityID;
                    else
                        Messages.Add(
                            new ExportImportMessage(
                                "Demo Entity for Template '" + t.Name + "' could not be found. (Guid: " + DemoEntityGuid +
                                ")", ExportImportMessage.MessageTypes.Warning));

                }

                t.Script = template.Attribute("Script").Value;
                t.IsFile = Boolean.Parse(template.Attribute("IsFile").Value);
                t.Type = template.Attribute("Type").Value;
                t.IsHidden = Boolean.Parse(template.Attribute("IsHidden").Value);
                t.Location = template.Attribute("Location").Value;

                if (template.Attribute("UseForList") != null)
                    t.UseForList = Boolean.Parse(template.Attribute("UseForList").Value);

                t.PortalID = PortalSettings.Current.PortalId;
                Sexy.TemplateContext.AddTemplate(t);

                foreach (XElement Entity in template.Elements("Entity"))
                    ImportEntity(Entity, Messages, Sexy.SexyContentTemplateAssignmentObjectTypeID, t.TemplateID);

                Messages.Add(new ExportImportMessage("Template '" + t.Name + "' successfully imported.",
                                                     ExportImportMessage.MessageTypes.Information));
            }

            Sexy.TemplateContext.SaveChanges();
        }

        #endregion

        /// <summary>
        /// Imports an Entity XElement
        /// </summary>
        /// <param name="entity">The XElement with the Entity</param>
        /// <param name="messages">The ExportImportMessage List that collects messages while importing</param>
        /// <param name="Change">The ChangeLog ChangeID</param>
        /// <param name="assignmentObjectTypeID"></param>
        /// <param name="keyNumber"></param>
        /// <returns></returns>
        private bool ImportEntity(XElement entity, List<ExportImportMessage> messages, int assignmentObjectTypeID, int? keyNumber = null)
        {
            var context = Sexy;
            var entityGuid = Guid.Parse(entity.Attribute("EntityGUID").Value);

            // Need a different Zone (new Context) when importing entities attached to fieldproperties or templates
            if (assignmentObjectTypeID == DataSource.AssignmentObjectTypeIdFieldProperties || assignmentObjectTypeID == Sexy.SexyContentTemplateAssignmentObjectTypeID)
            {
                context = new SexyContent(false, DataSource.DefaultZoneId);
                context.ContentContext.SetChangeLogId(Sexy.ContentContext.GetChangeLogId(HttpContext.Current.User.Identity.Name));
                entityGuid = Guid.NewGuid();
            }

            var attSet = context.ContentContext.GetAttributeSet(entity.Attribute("AttributeSetStaticName").Value);

            #region Return if entity already exists or content type does not exist
            if (context.ContentContext.EntityExists(entityGuid))
            {
                messages.Add(new ExportImportMessage("Error while importing Content Item:  The Entity already exists.", ExportImportMessage.MessageTypes.Warning));
                return false;
            }

            if (attSet == null)
            {
                messages.Add(new ExportImportMessage("Error while importing Content Item: Matching Content Type not found.", ExportImportMessage.MessageTypes.Warning));
                return false;
            }
            #endregion

            // Add Entity
            var newEntity = context.ContentContext.AddEntity(attSet.AttributeSetID, new Dictionary<string, string>() {}, null, keyNumber, assignmentObjectTypeID, 0, entityGuid);
            var targetDimensions = context.ContentContext.GetDimensionChildren("Culture");

            foreach (var att in attSet.AttributesInSets)
            {
                var values = entity.Elements("Value").Where(a => a.Attribute("Key").Value == att.Attribute.StaticName).ToList();
                var valuesToImport = new List<ImportValue>();

                if (!values.Any())
                {
                    messages.Add(new ExportImportMessage("Error while importing Content Item: No Value found for '" + att.Attribute.StaticName + "' in XML.", ExportImportMessage.MessageTypes.Warning));
                    continue;
                }

                // ToDo: Import / Export Localization

                // Process each target's language
                foreach (var dimension in targetDimensions)
                {
                    var exactMatchSourceDimension = sourceDimensions.FirstOrDefault(p => p.Attribute("ExternalKey").Value == dimension.ExternalKey);
                    var unExactMatchSourceDimensions = sourceDimensions.Where(p => p.Attribute("ExternalKey").Value.StartsWith(dimension.ExternalKey.Substring(0, 3)));

                    XElement value = null;

                    if (exactMatchSourceDimension != null)
                        value = values.FirstOrDefault(p => p.Elements("Dimension").Any(d => d.Attribute("DimensionID").Value == exactMatchSourceDimension.Attribute("DimensionID").Value));

                    if (value == null)
                        value = values.FirstOrDefault(p => p.Elements("Dimension").Any(d => unExactMatchSourceDimensions.Any(u => u.Attribute("DimensionID").Value == d.Attribute("DimensionID").Value)));


                    // Try to find exact match value
                    //var value = values.Where(v => v.)


                    // If value was not found & current dimension is default language, assign source's master language value
                    if (value == null && dimension.ExternalKey == PortalSettings.Current.CultureCode)
                    {
                        value = values.FirstOrDefault(p => p.Elements("Dimension"))
                    }

                }

                //var targetDefaultLanguageValue = values.FirstOrDefault(p => p.Elements("Dimension").Any())









                foreach (var value in values)
                {


                    #region Temporary
                    // ToDo: Fix
                    // Special case for Template ContentTypes
                    //if (attSet.StaticName == "2SexyContent-Template-ContentTypes")
                    //{
                    //    switch (att.Attribute.StaticName)
                    //    {
                    //        case "ContentTypeID":
                    //            var attributeSet = Sexy.ContentContext.GetAllAttributeSets().FirstOrDefault(a => a.AttributeSetID == int.Parse(x.Attribute("Value").Value));
                    //            x.Attribute("Value").SetValue(attributeSet != null ? attributeSet.StaticName : String.Empty);
                    //            break;
                    //        case "DemoEntityID":
                    //            var entityID = int.Parse(x.Attribute("Value").Value);
                    //            var demoEntity = Sexy.ContentContext.Entities.FirstOrDefault(e => e.EntityID == entityID);
                    //            x.Attribute("Value").SetValue(demoEntity != null ? demoEntity.EntityGUID.ToString() : String.Empty);
                    //            break;
                    //    }
                    //}

                    //var addedValue = new EavValue()
                    //    {
                    //        AttributeID = att.AttributeID,
                    //        EntityID = newEntity.EntityID,
                    //        Value = value.Attribute("Value").Value,
                    //        ChangeLogIDCreated = importChangeId
                    //    };

                    //Sexy.ContentContext.AddToValues(addedValue);

                    //foreach (var dimension in value.Elements("Dimension"))
                    //{
                    //    List<Dimension> dimensionsToAdd = new List<Dimension>();

                    //    var currentDimension =
                    //        Sexy.ContentContext.GetDimensionChildren("Culture")
                    //            .FirstOrDefault(
                    //                d =>
                    //                d.ExternalKey ==
                    //                sourceDimensions.FirstOrDefault(
                    //                    p =>
                    //                    p.Attribute("DimensionID").Value == dimension.Attribute("DimensionID").Value)
                    //                                .Attribute("ExternalKey")
                    //                                .Value);

                    //    if (currentDimension != null)
                    //        dimensionsToAdd.Add(currentDimension);

                    //    // Add current default dimension if current dimension was default language and default language of this portal is different
                    //    if (sourceDefaultDimension != PortalSettings.Current.CultureCode &&
                    //        sourceDefaultDimension == currentDimension.ExternalKey)
                    //    {
                    //        var currentDefaultDimension =
                    //            Sexy.ContentContext.GetDimensionChildren("Culture")
                    //                .FirstOrDefault(p => p.ExternalKey == PortalSettings.Current.CultureCode);

                    //        if (currentDefaultDimension != null)
                    //            dimensionsToAdd.Add(currentDefaultDimension);
                    //    }

                    //    // Add dimensions to value
                    //    foreach (var dim in dimensionsToAdd)
                    //    {
                    //        Sexy.ContentContext.AddToValuesDimensions(new ValueDimension()
                    //            {
                    //                Value = addedValue,
                    //                DimensionID = dim.DimensionID,
                    //                ReadOnly = Boolean.Parse(dimension.Attribute("ReadOnly").Value)
                    //            });
                    //    }
                    //}
                    #endregion
                }


            }

            Sexy.ContentContext.SaveChanges();

            // Loop fields in XML to show Message if additional fields are in the xml
            // ToDo: Maybe do this again in the future?
            //foreach (string Key in Entity.Elements("Value").Select(p => p.Attribute("Key")))
            //{
            //    if(!Values.Contains(Key))
            //        Messages.Add(new ExportImportMessage("Warning: Attribute '" + Key + "' found in XML but not in Content Type.", ExportImportMessage.MessageTypes.Warning));
            //}





            //Context.ContentContext.AddEntity(AttSet.AttributeSetID, Values, null, KeyNumber, AssignmentObjectTypeID, 0, EntityGuid);

            messages.Add(new ExportImportMessage("Content Item for '" + attSet.Name + "' successfully imported.",
                                                 ExportImportMessage.MessageTypes.Information));

            return true;
        }

    }
}