﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Web;
using DotNetNuke.Entities.Portals;
using DotNetNuke.Services.Exceptions;
using ICSharpCode.SharpZipLib.Zip;

namespace ToSic.SexyContent.ImportExport
{
    public class ZipImport
    {

        /// <summary>
        /// Imports a ZIP file (from stream)
        /// </summary>
        /// <param name="ZipStream"></param>
        /// <param name="Server"></param>
        /// <param name="PortalSettings"></param>
        /// <param name="Messages"></param>
        /// <returns></returns>
        public bool ImportZip(Stream ZipStream, HttpServerUtility Server, PortalSettings PortalSettings, List<ExportImportMessage> Messages)
        {
            if (Messages == null)
                Messages = new List<ExportImportMessage>();

            string TemporaryDirectory = Server.MapPath(Path.Combine(SexyContent.TemporaryDirectory, System.Guid.NewGuid().ToString()));
            var success = true;

            try
            {
                if (!Directory.Exists(TemporaryDirectory))
                    Directory.CreateDirectory(TemporaryDirectory);

                // Extract ZIP archive to the temporary folder
                ExtractZipFile(ZipStream, TemporaryDirectory);

                string CurrentWorkingDir = TemporaryDirectory;
                string[] BaseDirectories = Directory.GetDirectories(CurrentWorkingDir);

                // Loop through each root-folder. For now only contains the "Apps" folder.
                foreach (string DirectoryPath in BaseDirectories)
                {
                    switch (Path.GetFileName(DirectoryPath))
                    {
                        // Handle the App folder
                        case "Apps":
                            CurrentWorkingDir = Path.Combine(CurrentWorkingDir, "Apps");

                            // Loop through each app directory
                            foreach (string AppDirectory in Directory.GetDirectories(CurrentWorkingDir))
                            {
                                // Copy all files in 2sexy folder to (portal file system) 2sexy folder
                                string TemplateRoot = Server.MapPath(SexyContent.GetTemplatePathRoot(SexyContent.TemplateLocations.PortalFileSystem, PortalSettings));
                                string AppTemplateRoot = Path.Combine(AppDirectory, "2sexy");
                                if (Directory.Exists(AppTemplateRoot))
                                    CopyAllFiles(AppTemplateRoot, TemplateRoot, false, Messages);

                                // Import each XML file which is in the current App folder
                                foreach (string XmlFileName in Directory.GetFiles(AppDirectory, "*.xml"))
                                {
                                    string FileContents = File.ReadAllText(Path.Combine(AppDirectory, XmlFileName));
                                    new XmlImport().ImportXml(FileContents, Messages);
                                }
                            }

                            // Reset CurrentWorkingDir
                            CurrentWorkingDir = TemporaryDirectory;
                            break;
                    }
                }
            }
            catch (Exception e)
            {
                // Add error message and return false
                Messages.Add(new ExportImportMessage("Could not import the package. " + e.Message, ExportImportMessage.MessageTypes.Error));
                Exceptions.LogException(e);
                success = false;
            }
            finally
            {
                // Finally delete the temporary directory
                Directory.Delete(TemporaryDirectory, true);
            }

            return success;
        }

        #region Zip Import Helpers

        /// <summary>
        /// Copy all files from SourceFolder to DestinationFolder
        /// </summary>
        /// <param name="SourceFolder"></param>
        /// <param name="DestinationFolder"></param>
        /// <param name="OverwriteFiles"></param>
        private void CopyAllFiles(string SourceFolder, string DestinationFolder, Boolean OverwriteFiles, List<ExportImportMessage> Messages)
        {
            var FileList = from f in Directory.GetFiles(SourceFolder, "*.*", SearchOption.AllDirectories)
                           select f;

            foreach (string File in FileList)
            {
                string RelativeFilePath = File.Replace(SourceFolder, "");
                string DestinationFilePath = String.Format("{0}{1}{2}",
                DestinationFolder, Path.DirectorySeparatorChar, RelativeFilePath);

                if (!Directory.Exists(Path.GetDirectoryName(DestinationFilePath)))
                {
                    Directory.CreateDirectory(Path.GetDirectoryName(DestinationFilePath));
                }

                if (!System.IO.File.Exists(DestinationFilePath))
                    System.IO.File.Copy(File, DestinationFilePath, OverwriteFiles);
                else
                    Messages.Add(new ExportImportMessage("File '" + DestinationFilePath + "' not copied because it already exists", ExportImportMessage.MessageTypes.Warning));

            }
        }

        /// <summary>
        /// Extracts a Zip (as Stream) to the given OutFolder directory.
        /// </summary>
        /// <param name="ZipStream"></param>
        /// <param name="OutFolder"></param>
        private void ExtractZipFile(Stream ZipStream, string OutFolder)
        {
            ZipFile File = new ZipFile(ZipStream);

            try
            {
                foreach (ZipEntry Entry in File)
                {
                    if (Entry.IsDirectory)
                        continue;
                    string FileName = Entry.Name;

                    Stream EntryStream = File.GetInputStream(Entry);

                    string FullPath = Path.Combine(OutFolder, FileName);
                    string DirectoryName = Path.GetDirectoryName(FullPath);
                    if (!String.IsNullOrEmpty(DirectoryName))
                        Directory.CreateDirectory(DirectoryName);

                    // Unzip File in buffered chunks
                    using (FileStream StreamWriter = System.IO.File.Create(FullPath))
                    {
                        EntryStream.CopyTo(StreamWriter, 4096);
                    }
                }
            }
            finally
            {
                if (File != null)
                    File.Close();
            }
        }

        #endregion
    }
}