﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Web;
using System.Web.UI;
using DotNetNuke.Entities.Portals;
using DotNetNuke.Security.Roles;
using DotNetNuke.Services.Search;
using ToSic.Eav;
using DotNetNuke.Common.Utilities;
using DotNetNuke.Services.Localization;
using DotNetNuke.Entities.Modules;
using System.Reflection;
using System.Web.UI.HtmlControls;
using ToSic.Eav.DataSources;
using ToSic.SexyContent;

namespace ToSic.SexyContent
{
    /// <summary>
    /// Business Layer
    /// Centralizes all global constants and business logic
    /// </summary>
    public class SexyContent : ISearchable
    {
        #region Constants

        public const string ModuleVersion = "05.03.02";
        public const string TemplateID = "TemplateID";
        public const string ContentGroupIDString = "ContentGroupID";
        public const string SettingsPublishDataSource = "ToSic_SexyContent_PublishDataSource";
        public const string SettingsPublishDataSourceStreams = "ToSic_SexyContent_PublishDataSource_Streams";
        public const string ContentGroupItemIDString = "ContentGroupItemID";
        public const string SortOrderString = "SortOrder";
        public const string EntityID = "EntityID";
        public const string AttributeSetIDString = "AttributeSetID";
        public const string RazorC = "C# Razor";
        public const string RazorVB = "VB Razor";
        public const string TokenReplace = "Token";

        public static class ControlKeys
        {
            public const string View = "";
            public const string Settings = "settings";
            public const string SettingsWrapper = "settingswrapper";
            public const string EavManagement = "eavmanagement";
            public const string EditContentGroup = "editcontentgroup";
            public const string EditTemplate = "edittemplate";
            public const string ManageTemplates = "managetemplates";
            public const string EditList = "editlist";
            public const string TemplateHelp = "templatehelp";
            public const string Export = "export";
            public const string Import = "import";
            public const string EditTemplateFile = "edittemplatefile";
            public const string AddItem = "additem";
            public const string EditTemplateDefaults = "edittemplatedefaults";
            public const string GettingStarted = "gettingstarted";
            public const string PortalConfiguration = "portalconfiguration";
            public const string EditDataSource = "editdatasource";
        }

        public const string PortalHostDirectory = "~/Portals/_default/";
        public const string LocationIDCurrentPortal = "Portal File System";
        public const string TemplateFolder = "2sexy";
        public const string PortalSettingsPrefix = "ToSIC_SexyContent_";

        /// <summary>
        /// Collection of Template Locations
        /// </summary>
        public class TemplateLocations
        {
            public const string PortalFileSystem = "Portal File System";
            public const string HostFileSystem = "Host File System";
        }

        public const string WebConfigTemplatePath = "~/DesktopModules/ToSIC_SexyContent/WebConfigTemplate.config";
        public const string WebConfigFileName = "web.config";
        public const string SexyContentGroupName = "2SexyContent Designers";
        public const string AttributeSetScope = "2SexyContent";
        public const string TemplateMetaDataAttributeSetStaticName = "2SexyContent-Template-Metadata";
        public const string TemplateContentTypesAttributeSetStaticName = "2SexyContent-Template-ContentTypes";
        public const string TemporaryDirectory = "~/DesktopModules/ToSIC_SexyContent/Temporary";

        #endregion

        #region Properties

        /// <summary>
        /// The Content Data Context
        /// </summary>
        private EavContext _ContentContext;

        public EavContext ContentContext
        {
            get
            {
                if (_ContentContext == null)
                    _ContentContext = EavContext.Instance(ContextZoneID.HasValue ? ContextZoneID : DataSource.DefaultZoneId, null);
                return _ContentContext;
            }
        }

        private int? ContextZoneID { get; set; }

        /// <summary>
        /// The Template Data Context
        /// </summary>
        public SexyContentContext TemplateContext { get; internal set; }

        /// <summary>
        /// The EAV DataSource
        /// </summary>
        public ToSic.Eav.DataSources.IDataSource SexyDataSource { get; internal set; }

        #endregion

        #region AssignmentObjectType Lookups
        /// <summary>
        /// Returns the AssignmentObjectTypeID for all 2SexyContent stuff
        /// </summary>
        public int SexyContentAssignmentObjectTypeID
        {
            get
            {
                return DataSource.GetCache(DataSource.DefaultZoneId, DataSource.MetaDataAppId).GetAssignmentObjectTypeId("2SexyContent");
            }
        }

        /// <summary>
        /// Returns the AssignmentObjectTypeID for 2SexyContent Templates
        /// </summary>        
        public int SexyContentTemplateAssignmentObjectTypeID
        {
            get
            {
                return DataSource.GetCache(DataSource.DefaultZoneId, DataSource.MetaDataAppId).GetAssignmentObjectTypeId("2SexyContent-Template");
            }
        }

        #endregion

        #region SexyContent Constructor

        /// <summary>
        /// Instanciates Content and Template-Contexts
        /// </summary>
        public SexyContent(bool EnableCaching = true, int? ZoneID = null, int? AppID = new int?())
        {
            // Only disable caching of templates and contentgroupitems
            // if AppSetting "ToSIC_SexyContent_EnableCaching" is disabled
            if(EnableCaching)
            {
                var CachingSetting = System.Configuration.ConfigurationManager.AppSettings["ToSIC_SexyContent_EnableCaching"];
                if (!String.IsNullOrEmpty(CachingSetting) && CachingSetting.ToLower() == "false")
                    EnableCaching = false;
            }

            // Get Entity Framework ConnectionString
            var entityBuilder = new System.Data.EntityClient.EntityConnectionStringBuilder();
            entityBuilder.ProviderConnectionString = Config.GetConnectionString();
            entityBuilder.Metadata = @"res://ToSic.SexyContent/SexyContent.SexyContentContext.csdl|
								res://ToSic.SexyContent/SexyContent.SexyContentContext.ssdl|
								res://ToSic.SexyContent/SexyContent.SexyContentContext.msl";
            entityBuilder.Provider = "System.Data.SqlClient";

            // Create TemplateContext
            TemplateContext = new SexyContentContext(entityBuilder.ToString(), EnableCaching);

            if (!ZoneID.HasValue && PortalSettings.Current != null)
                ZoneID = GetZoneID(PortalSettings.Current.PortalId);

            this.ContextZoneID = ZoneID;

            // Set EAV Connection String
            ToSic.Eav.Configuration.SetConnectionString("SiteSqlServer");

            // Set Properties on ContentContext
            ContentContext.UserName = (HttpContext.Current == null) ? "Internal" : HttpContext.Current.User.Identity.Name;
            if (AppID.HasValue)
                ContentContext.AppId = AppID.Value;
        }

        /// <summary>
        /// Constructor overload for DotNetNuke
        /// (BusinessControllerClass needs parameterless constructor)
        /// </summary>
        public SexyContent() : this(true, new int?(), new int?())
        {
        }
        #endregion

        #region Template File Handling
        /// <summary>
        /// Returns all template files in the template folder.
        /// </summary>
        /// <param name="Server">The HttpServerUtility for MapPath</param>
        /// <param name="PortalSettings"></param>
        /// <param name="TemplateType"></param>
        /// <param name="TemplateLocation"></param>
        /// <param name="ControlPath">The PortalModuleBase.ControlPath property</param>
        /// <returns></returns>
        public IEnumerable<string> GetTemplateFiles(HttpServerUtility Server, PortalSettings PortalSettings, string TemplateType, string TemplateLocation)
        {
            string TemplatePathRootMapPath = Server.MapPath(GetTemplatePathRoot(TemplateLocation, PortalSettings));
            DirectoryInfo Directory = new DirectoryInfo(TemplatePathRootMapPath);

            EnsureTemplateFolderExists(Server, Directory);

            // Filter the files according to type
            string FileFilter = "*.html";
            switch (TemplateType)
            {
                case RazorC:
                    FileFilter = "*.cshtml";
                    break;
                case RazorVB:
                    FileFilter = "*.vbhtml";
                    break;
                case TokenReplace:
                    FileFilter = "*.html";
                    break;
            }

            FileInfo[] Files = Directory.GetFiles(FileFilter, SearchOption.AllDirectories);
            return (from d in Files where d.Name != SexyContent.WebConfigFileName select (d.FullName).Replace(TemplatePathRootMapPath, "").Replace('\\','/'));
        }

        /// <summary>
        /// Creates a template file if it does not already exists, and uses a default text to insert.
        /// </summary>
        /// <param name="Name"></param>
        /// <param name="Template"></param>
        /// <param name="PortalSettings"></param>
        /// <param name="Server"></param>
        /// <param name="Contents"></param>
        public void CreateTemplateFileIfNotExists(string Name, Template Template, PortalSettings PortalSettings, HttpServerUtility Server, string Contents = "")
        {
            if (Template.Type == RazorC)
            {
                if (!Name.StartsWith("_"))
                    Name = "_" + Name;
                if (Path.GetExtension(Name) != ".cshtml")
                    Name += ".cshtml";
            }
            else if (Template.Type == RazorVB)
            {
                if (!Name.StartsWith("_"))
                    Name = "_" + Name;
                if (Path.GetExtension(Name) != ".vbhtml")
                    Name += ".vbhtml";
            }
            else if (Template.Type == TokenReplace)
            {
                if (Path.GetExtension(Name) != ".html")
                    Name += ".html";
            }

            Template.Path = System.Text.RegularExpressions.Regex.Replace(Name, @"[?:\/*""<>|]", "");
            var TemplatePath = Server.MapPath(Template.GetTemplatePath(PortalSettings));

            if (!File.Exists(TemplatePath))
            {
                StreamWriter Stream = new StreamWriter(File.Create(TemplatePath));
                Stream.Write(Contents);
                Stream.Flush();
                Stream.Close();
            }
        }

        /// <summary>
        /// Creates a directory and copies the needed web.config for razor files
        /// if the directory does not exist.
        /// </summary>
        /// <param name="Server"></param>
        /// <param name="Directory"></param>
        /// <param name="ControlPath"></param>
        private void EnsureTemplateFolderExists(HttpServerUtility Server, DirectoryInfo Directory)
        {
            if (!Directory.Exists)
            {
                Directory.Create();
                File.Copy(Server.MapPath(WebConfigTemplatePath), Directory.FullName + WebConfigFileName);
            }
        }

        /// <summary>
        /// Configures a Portal (Creates a 2sexy Folder containing a web.config File)
        /// </summary>
        /// <param name="Server"></param>
        /// <param name="PortalSettings"></param>
        /// <param name="ControlPath"></param>
        public void ConfigurePortal(HttpServerUtility Server, PortalSettings PortalSettings)
        {
            string TemplatePathRootMapPath = Server.MapPath(GetTemplatePathRoot(TemplateLocations.PortalFileSystem, PortalSettings));
            DirectoryInfo Directory = new DirectoryInfo(TemplatePathRootMapPath);
            EnsureTemplateFolderExists(Server, Directory);
        }

        /// <summary>
        /// Returns true if the Portal HomeDirectory Contains the 2sexy Folder and this folder contains the web.config
        /// </summary>
        /// <param name="Server"></param>
        /// <param name="ControlPath"></param>
        /// <param name="PortalSettings"></param>
        /// <returns></returns>
        public bool PortalIsConfigured(HttpServerUtility Server, string ControlPath, PortalSettings PortalSettings)
        {
            DirectoryInfo TemplateDirectory = new DirectoryInfo(Server.MapPath(GetTemplatePathRoot(TemplateLocations.PortalFileSystem, PortalSettings)));
            FileInfo WebConfigTemplate = new FileInfo(TemplateDirectory.FullName + SexyContent.WebConfigFileName);
            return TemplateDirectory.Exists && WebConfigTemplate.Exists;
        }

        /// <summary>
        /// Returns the location, where Templates are stored
        /// Must be static because used in Template.cs
        /// </summary>
        /// <param name="LocationID"></param>
        /// <param name="PortalSettings"></param>
        /// <returns></returns>
        public static string GetTemplatePathRoot(string LocationID, DotNetNuke.Entities.Portals.PortalSettings PortalSettings)
        {
            string RootFolder = (LocationID == LocationIDCurrentPortal ? PortalSettings.HomeDirectory : PortalHostDirectory);
            RootFolder += TemplateFolder + "/";
            return RootFolder;
        }

        #endregion

        #region Template Configuration

        public void CreateOrUpdateTemplateDefault(int TemplateID, string ItemType, int? ContentTypeID, int? DemoEntityID)
        {
            var DefaultAppContext = new SexyContent(true, DataSource.DefaultZoneId);
            var AttributeSetID = DefaultAppContext.ContentContext.GetAttributeSet(TemplateContentTypesAttributeSetStaticName).AttributeSetID;
            var Entities = DefaultAppContext.ContentContext.GetEntities(SexyContentTemplateAssignmentObjectTypeID, TemplateID, null, null);

            var Values = new OrderedDictionary()
                                {
                                    {"ItemType", ItemType},
                                    {"ContentTypeID", ContentTypeID.HasValue ? ContentTypeID.Value.ToString() : "0" },
                                    {"DemoEntityID", DemoEntityID.HasValue ? DemoEntityID.Value.ToString() : "0" }
                                };

            var ExistingEntity =
                Entities.FirstOrDefault(
                    p => p.Values.Any(v => v.Attribute.StaticName == "ItemType" && v.Value == ItemType));

            if (ExistingEntity != null)
                DefaultAppContext.ContentContext.UpdateEntity(ExistingEntity.EntityID, Values);
            else
                DefaultAppContext.ContentContext.AddEntity(AttributeSetID, Values, null, TemplateID, SexyContentTemplateAssignmentObjectTypeID, 0);
        }

        public TemplateDefault GetTemplateDefault(int TemplateID, ContentGroupItemType ItemType)
        {
            return GetTemplateDefaults(TemplateID).FirstOrDefault(t => t.ItemType == ItemType);
        }

        public List<TemplateDefault> GetTemplateDefaults(int TemplateID)
        {
            var Result = new List<TemplateDefault>();
            var Entities = DataSource.GetMetaDataSource().GetAssignedEntities(SexyContentTemplateAssignmentObjectTypeID, TemplateID, TemplateContentTypesAttributeSetStaticName);

            // Add TemplateDefault configured directly in Template
            var Template = TemplateContext.GetTemplate(TemplateID);
            Result.Add(new TemplateDefault() { ContentTypeID = Template.AttributeSetID, DemoEntityID = Template.DemoEntityID, ItemType = ContentGroupItemType.Content });

            Result.AddRange(Entities.Select(e => new TemplateDefault()
                {
                    ItemType = (ContentGroupItemType) Enum.Parse(typeof (ContentGroupItemType), (string) e.Attributes["ItemType"][0]),
                    ContentTypeID = e.Attributes.ContainsKey("ContentTypeID") && (decimal) e.Attributes["ContentTypeID"][0] != 0 ? Convert.ToInt32((decimal) e.Attributes["ContentTypeID"][0]) : new int?(),
                    DemoEntityID = e.Attributes.ContainsKey("DemoEntityID") && (decimal) e.Attributes["DemoEntityID"][0] != 0 ? Convert.ToInt32((decimal) e.Attributes["DemoEntityID"][0]) : new int?()
                }));

            return Result;
        }

        public List<Template> GetCompatibleTemplates(int PortalID, int ContentGroupID)
        {
            var ContentGroupItems = TemplateContext.GetContentGroupItems(ContentGroupID).ToList();
            List<Template> CompatibleTemplates;

            // Prepare some variables
            var List = ContentGroupItems.Count(p => p.ItemType == ContentGroupItemType.Content) > 1;
            var CurrentTemplate = ContentGroupItems.First().Template;
            var CurrentDefaults = GetTemplateDefaults(CurrentTemplate.TemplateID);

            CompatibleTemplates = TemplateContext.GetTemplates(PortalID).Where(t => t.UseForList || !List).ToList();
            CompatibleTemplates = CompatibleTemplates.Where(c =>
                AreTemplateDefaultsCompatible(ContentGroupItems, CurrentDefaults, GetTemplateDefaults(c.TemplateID))).ToList();

            return CompatibleTemplates;
        }

        private bool AreTemplateDefaultsCompatible(List<ContentGroupItem> ContentGroupItems, IEnumerable<TemplateDefault> Current, IEnumerable<TemplateDefault> New)
        {
            return Current.All(c => IsTemplateDefaultCompatible(ContentGroupItems, c, New.FirstOrDefault(d => d.ItemType == c.ItemType)));
        }

        private bool IsTemplateDefaultCompatible(List<ContentGroupItem> ContentGroupItems, TemplateDefault Current, TemplateDefault New)
        {
            if (Current == null || New == null)
                return false;

            if (!ContentGroupItems.Any(c => c.ItemType == Current.ItemType && c.EntityID.HasValue))
                return true;

            return (Current.ItemType == New.ItemType && Current.ContentTypeID == New.ContentTypeID);
        }

        #endregion

        #region User Management
        /// <summary>
        /// Returns true if a DotNetNuke User Group "SexyContent Designers" exists and contains at minumum one user
        /// </summary>
        /// <param name="PortalID"></param>
        /// <returns></returns>
        public bool SexyContentDesignersGroupConfigured(int PortalID)
        {
            DotNetNuke.Security.Roles.RoleController RoleControl = new DotNetNuke.Security.Roles.RoleController();
            RoleInfo Role = RoleControl.GetRoleByName(PortalID, SexyContentGroupName);
            if (Role != null)
            {
                System.Collections.ArrayList t = RoleControl.GetUsersByRoleName(PortalID, SexyContentGroupName);
                if (t.Count > 0)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Returns true if a user is in the SexyContent Designers group
        /// </summary>
        /// <param name="User"></param>
        /// <returns></returns>
        public bool IsInSexyContentDesignersGroup(DotNetNuke.Entities.Users.UserInfo User)
        {
            return User.IsInRole(SexyContentGroupName);
        }
        #endregion

        #region Preparation of 2Sexy Elements

        public ToSic.Eav.DataSources.IDataSource GetInitialDataSource()
        {
            var zoneId = PortalSettings.Current == null ? this.ContextZoneID : GetZoneID(PortalSettings.Current.PortalId);
            return DataSource.GetInitialDataSource(zoneId: zoneId);
        }

        /// <summary>
        /// Get a list of ContentElements by ModuleId or ContentGroupID
        /// </summary>
        /// <returns></returns>
        public List<Element> GetContentElements(int ModuleID, string LanguageName, int? ContentGroupID, int PortalId)
        {
            return GetElements(ModuleID, ContentGroupID, ContentGroupItemType.Content, ContentGroupItemType.Presentation, LanguageName, PortalId);
        }

        public Element GetListElement(int ModuleID, string LanguageName, int? ContentGroupID, int PortalId)
        {
            var ListElement = GetElements(ModuleID, ContentGroupID, ContentGroupItemType.ListContent, ContentGroupItemType.ListPresentation, LanguageName, PortalId).FirstOrDefault();
            return ListElement;
        }

        private List<Element> GetElements(int ModuleID, int? ContentGroupID, ContentGroupItemType ContentItemType, ContentGroupItemType PresentationItemType, string LanguageName, int PortalId)
        {
            if (!ContentGroupID.HasValue)
                ContentGroupID = GetContentGroupIDFromModule(ModuleID);


            IEnumerable<ContentGroupItem> ItemsQuery = TemplateContext.GetContentGroupItems(ContentGroupID.Value);
            List<ContentGroupItem> Items = ItemsQuery.Where(c => c.Type == ContentItemType.ToString("F") || c.Type == PresentationItemType.ToString("F")).ToList();
            List<Element> ElementList = new List<Element>();

            if (ItemsQuery.Any(c => c.TemplateID.HasValue))
            {
                var Defaults = GetTemplateDefaults(ItemsQuery.First().TemplateID.Value);

                // Get List of Entities
                List<int> Identities = (from c in Items where c.EntityID.HasValue && c.EntityID.Value > 0 select c.EntityID.Value).ToList();
                // Add Demo Entities
                Identities.AddRange(Defaults.Where(d => d.DemoEntityID.HasValue).Select(d => d.DemoEntityID.Value));
                // Prepare Dimension List (Languages)
                var DimensionIds = new[] { LanguageName };
                // Load all Entities to list
                var InitialSource = GetInitialDataSource();
                var EntityDataSource = DataSource.GetDataSource("ToSic.Eav.DataSources.EntityIdFilter", this.GetZoneID(PortalId), null, InitialSource);
                ((EntityIdFilter)EntityDataSource).Configuration["EntityIds"] = String.Join(",", Identities.ToArray());
                SexyDataSource = DataSource.GetDataSource("ToSic.Eav.DataSources.PassThrough", this.GetZoneID(PortalId), null, EntityDataSource);
                var Entities = SexyDataSource.Out["Default"].List;// ContentContext.GetEntityModel(Identities);

                // If no Content Elements exist and type is List, add a ContentGroupItem to List (not to DB)
                if (ContentItemType == ContentGroupItemType.ListContent && Items.All(p => p.ItemType != ContentGroupItemType.ListContent))
                {
                    var ListContentDefault = Defaults.FirstOrDefault(d => d.ItemType == ContentGroupItemType.ListContent);
                    var TemplateID = ItemsQuery.First().TemplateID.Value;

                    Items.Add(new ContentGroupItem()
                    {
                        ContentGroupID = ContentGroupID.Value,
                        ContentGroupItemID = -1,
                        EntityID = ListContentDefault != null ? ListContentDefault.DemoEntityID : new int?(),
                        SortOrder = -1,
                        SysCreated = DateTime.Now,
                        SysCreatedBy = -1,
                        TemplateID = TemplateID,
                        Type = ContentGroupItemType.ListContent.ToString("F")
                    });
                }

                // Transform to list of Elements
                ElementList = (from c in Items
                               where c.ItemType == ContentItemType
                               select new Element
                               {
                                   ID = c.ContentGroupItemID,
                                   EntityID = c.EntityID,
                                   TemplateID = c.TemplateID,
                                   Content = c.EntityID.HasValue ? new DynamicEntity(Entities[c.EntityID.Value], DimensionIds) :
                                       Defaults.Where(d => d.ItemType == ContentItemType && d.DemoEntityID.HasValue)
                                       .Select(d => new DynamicEntity(Entities[d.DemoEntityID.Value], DimensionIds)).FirstOrDefault(),
                                   // Get Presentation object - Take Default if it does not exist
                                   Presentation = (from p in Items
                                                   where p.SortOrder == c.SortOrder && p.ItemType == PresentationItemType && p.EntityID.HasValue
                                                   select new DynamicEntity(Entities[p.EntityID.Value], DimensionIds)).FirstOrDefault() ??
                                                   (from d in Defaults
                                                    where d.ItemType == PresentationItemType && d.DemoEntityID.HasValue
                                                    select new DynamicEntity(Entities[d.DemoEntityID.Value], DimensionIds)).FirstOrDefault()
                                                   ,
                                   GroupID = c.ContentGroupID,
                                   SortOrder = c.SortOrder
                               }).ToList();
            }

            return ElementList;
        }

        /// <summary>
        /// Returns the ContentGroupID for a module.
        /// If it is not set, the ModuleID will set as ContentGroupID.
        /// </summary>
        /// <param name="ModuleID"></param>
        /// <returns></returns>
        public int GetContentGroupIDFromModule(int ModuleID)
        {
            ModuleController ModuleControl = new ModuleController();

            Hashtable Settings = ModuleControl.GetModuleSettings(ModuleID);

            if (Settings[ContentGroupIDString] == null)
                ModuleControl.UpdateModuleSetting(ModuleID, ContentGroupIDString, ModuleID.ToString());

            Settings = ModuleControl.GetModuleSettings(ModuleID);

            return int.Parse(Settings[SexyContent.ContentGroupIDString].ToString());
        }
        #endregion

        #region URL Handling / Toolbar

        /// <summary>
        /// Get Toolbar HTML for given Element
        /// </summary>
        /// <param name="Element"></param>
        /// <param name="HostingModule"></param>
        /// <param name="LocalResourcesPath"></param>
        /// <param name="ListEnabled"></param>
        /// <returns></returns>
        public string GetElementToolbar(int ContentGroupID, int SortOrder, int ContentGroupItemID, int ModuleId, string LocalResourcesPath, bool ListEnabled, Control ParentControl, string ReturnUrl)
        {
            string EditLink = GetElementEditLink(ContentGroupID, SortOrder, ModuleId, PortalSettings.Current.ActiveTab.TabID, ReturnUrl);

            string Toolbar = "<ul class=\"sc-menu\">";

            Toolbar += "<li><a class=\"sc-menu-edit\" href=\"" + UrlUtils.PopUpUrl(EditLink, null, PortalSettings.Current, false, false) + "\"><img src=\"" + ParentControl.ResolveClientUrl("~/DesktopModules/ToSIC_SexyContent/Images/Edit.png") + "\" /></a></li>";

            if (ListEnabled)
                Toolbar += "<li><a class=\"sc-menu-add\" href=\"javascript:void(0);\" onclick='AddContentGroupItem(this, \"" + ContentGroupItemID.ToString() + "\");'><img src=\"" + ParentControl.ResolveClientUrl("~/DesktopModules/ToSIC_SexyContent/Images/Add.png") + "\" /></a></li>";

            Toolbar += "</ul>";
            return Toolbar;
        }

        /// <summary>
        /// Get the URL for editing MetaData
        /// </summary>
        /// <param name="TabID"></param>
        /// <param name="ModuleID"></param>
        /// <param name="ReturnUrl"></param>
        /// <param name="PortalSettings"></param>
        /// <param name="Control"></param>
        /// <param name="AttributeSetStaticName"></param>
        /// <param name="AssignmentObjectTypeID"></param>
        /// <param name="KeyNumber"></param>
        /// <returns></returns>
        public string GetMetaDataEditUrl(int TabID, int ModuleID, string ReturnUrl, PortalSettings PortalSettings, System.Web.UI.Control Control, string AttributeSetStaticName, int AssignmentObjectTypeID, int KeyNumber)
        {
            var DefaultAppContext = new SexyContent(true, DataSource.DefaultZoneId);
            ToSic.Eav.AttributeSet Set = DefaultAppContext.ContentContext.GetAttributeSet(AttributeSetStaticName);
            string NewItemUrl = UrlUtils.PopUpUrl(DotNetNuke.Common.Globals.NavigateURL(TabID, ControlKeys.EavManagement, "mid=" + ModuleID.ToString() + "&ManagementMode=NewItem&AttributeSetID=[AttributeSetID]&KeyNumber=[KeyNumber]&AssignmentObjectTypeID=[AssignmentObjectTypeID]&ReturnUrl=[ReturnUrl]"), Control, PortalSettings, false, true);
            string EditItemUrl = UrlUtils.PopUpUrl(DotNetNuke.Common.Globals.NavigateURL(TabID, ControlKeys.EavManagement, "mid=" + ModuleID.ToString() + "&ManagementMode=EditItem&EntityID=[EntityID]&ReturnUrl=[ReturnUrl]"), Control, PortalSettings, false, true);
            return ToSic.Eav.ManagementUI.Forms.GetItemFormUrl(KeyNumber, Set.AttributeSetID, AssignmentObjectTypeID, NewItemUrl, EditItemUrl, ReturnUrl);
        }

        /// <summary>
        /// Returns the Edit Link for given GroupID and SortOrder
        /// </summary>
        /// <param name="Element"></param>
        /// <param name="ModuleID"></param>
        /// <param name="TabID"></param>
        /// <param name="UserID"></param>
        /// <param name="ReturnUrl"></param>
        /// <returns></returns>
        public string GetElementEditLink(int ContentGroupID, int SortOrder, int ModuleID, int TabID, string ReturnUrl)
        {
            string EditUrl = DotNetNuke.Common.Globals.NavigateURL(TabID, ControlKeys.EditContentGroup, "mid", ModuleID.ToString(), SortOrderString, SortOrder.ToString(), ContentGroupIDString, ContentGroupID.ToString());
            EditUrl += (EditUrl.IndexOf("?") == -1 ? "?" : "&") + "popUp=true&ReturnUrl=" + HttpUtility.UrlEncode(ReturnUrl);

            // If Culture exists, add CultureDimension
            var LanguageID = GetCurrentLanguageID();
            if(LanguageID.HasValue)
                EditUrl += "&CultureDimension=" + LanguageID;

            return EditUrl;
        }

        public string GetElementSettingsLink(int ContentGroupItemID, int ModuleID, int TabID, string ReturnUrl)
        {
            string SettingsUrl = DotNetNuke.Common.Globals.NavigateURL(TabID, ControlKeys.SettingsWrapper, "mid", ModuleID.ToString(), ContentGroupItemIDString, ContentGroupItemID.ToString());
            SettingsUrl += (SettingsUrl.IndexOf("?") == -1 ? "?" : "&") + "popUp=true&ReturnUrl=" + HttpUtility.UrlEncode(ReturnUrl);
            return SettingsUrl;
        }

        public void AttachToolbarToElements(List<Element> Elements, int ModuleID, string LocalResourcesPath, bool ListEnabled, bool IsEditable, Control ParentControl, string ReturnUrl)
        {
            // Add Toolbar if neccessary, else add empty string to dictionary
            if(!IsEditable)
                Elements.Where(p => p.Content != null).ToList().ForEach(p => ((DynamicEntity)p.Content).Toolbar = "");
            else
                Elements.Where(p => p.Content != null).ToList().ForEach(p => ((DynamicEntity)p.Content).Toolbar = GetElementToolbar(p.GroupID, p.SortOrder, p.ID, ModuleID, LocalResourcesPath, ListEnabled, ParentControl, ReturnUrl));
        }

        #endregion

        #region ContentGroupItem Management

        /// <summary>
        /// Adds a ContentGroupItem to a specified ContentGroup and returns the created ContentGroupItem.
        /// </summary>
        /// <param name="ContentGroupID"></param>
        /// <param name="UserID"></param>
        /// <param name="TemplateID"></param>
        /// <param name="EntityID"></param>
        /// <param name="DestinationSortOrder"></param>
        /// <param name="AutoSave">Prevents from saving the item if false</param>
        /// <returns></returns>
        public ContentGroupItem AddContentGroupItem(int ContentGroupID, int UserID, int? TemplateID, int? EntityID, int? DestinationSortOrder, bool AutoSave, ContentGroupItemType ItemType, bool PreventSorting)
        {
            var Item = new ContentGroupItem()
            {
                ContentGroupID = ContentGroupID,
                SysCreatedBy = UserID,
                SysModifiedBy = UserID,
                SysCreated = DateTime.Now,
                SysModified = DateTime.Now,
                SortOrder = DestinationSortOrder.HasValue ? DestinationSortOrder.Value : 0,
                TemplateID = TemplateID,
                EntityID = EntityID,
                Type = ItemType.ToString()
            };

            TemplateContext.AddContentGroupItem(Item);

            if (AutoSave)
                TemplateContext.SaveChanges();

            if (!PreventSorting)
            {
                var GroupItems = TemplateContext.GetContentGroupItems(ContentGroupID);

                if (GroupItems.Any(p => p != Item))
                {
                    Item.SortOrder = GroupItems.Where(p => p != Item).Max(p => p.SortOrder) + 1;
                }

                if (DestinationSortOrder.HasValue)
                    TemplateContext.ReorderContentGroupItem(Item, DestinationSortOrder.Value, AutoSave);
            }

            if (AutoSave)
                TemplateContext.SaveChanges();

            return Item;
        }

        public void UpdateTemplateForGroup(int ContentGroupID, int? TemplateID, int UserID)
        {
            List<ContentGroupItem> Items = TemplateContext.GetContentGroupItems(ContentGroupID, ContentGroupItemType.Content).ToList();

            if (!Items.Any())
                Items.Add(AddContentGroupItem(ContentGroupID, UserID, null, null, null, true, ContentGroupItemType.Content, false));

            TemplateID = TemplateID == 0 ? null : TemplateID;
            Items.ForEach(p => p.TemplateID = TemplateID);

            TemplateContext.SaveChanges();
        }

        public IEnumerable<AttributeSet> GetAvailableAttributeSets()
        {
            return from c in ContentContext.GetAllAttributeSets()
                   where !c.Name.StartsWith("@")
                         && c.Scope == SexyContent.AttributeSetScope
                         && !c.ChangeLogIDDeleted.HasValue
                   select c;
        }

        public IEnumerable<AttributeSet> GetAvailableAttributeSetsForVisibleTemplates(int PortalId)
        {
            var AvailableTemplates = this.TemplateContext.GetVisibleTemplates(PortalId);
            return GetAvailableAttributeSets().Where(p => AvailableTemplates.Any(t => t.AttributeSetID == p.AttributeSetID)).OrderBy(p => p.Name);
        }

        /// <summary>
        /// Returns if any ContentItem with the TemplateID and ItemType specified is in use.
        /// </summary>
        /// <param name="TemplateID"></param>
        /// <param name="ItemType"></param>
        /// <returns></returns>
        public bool IsTemplateDefaultInUse(int TemplateID, ContentGroupItemType ItemType)
        {
            return TemplateContext.GetContentGroupItems().Any(c => c.TemplateID == TemplateID && c.ItemType == ItemType && c.EntityID.HasValue);
        }

        #endregion

        #region DNN Interface Members

        public SearchItemInfoCollection GetSearchItems(ModuleInfo ModInfo)
        {
            var ZoneID = GetZoneID(ModInfo.PortalID);

            // Need a new Context because PortalSettings.Current is null
            var Sexy = new SexyContent(true, ZoneID);

            var SearchItems = new SearchItemInfoCollection();

            var Elements = Sexy.GetContentElements(ModInfo.ModuleID, Sexy.GetCurrentLanguageName(), null, ModInfo.PortalID);
            Elements.Add(Sexy.GetListElement(ModInfo.ModuleID, Sexy.GetCurrentLanguageName(), null, ModInfo.PortalID));

            foreach (var Element in Elements)
            {
                if (Element != null && Element.EntityID.HasValue)
                {
                    var Attributes = ((DynamicEntity)Element.Content).Entity.Attributes;
                    string Content = String.Join(", ", Attributes.Select(x => x.Value[new string[] { Sexy.GetCurrentLanguageName() }]).Where(a => a != null).Select(a => StripHtmlAndHtmlDecode(a.ToString())).Where(x => !String.IsNullOrEmpty(x)));

                    var ContentGroupItem = Sexy.TemplateContext.GetContentGroupItem(Element.ID);
                    var PubDate = Sexy.ContentContext.GetValues(Element.EntityID.Value).Max(p => (DateTime?)p.ChangeLogCreated.Timestamp);

                    if (PubDate.HasValue)
                        SearchItems.Add(new SearchItemInfo(Element.Content.EntityTitle, Content, ContentGroupItem.SysCreatedBy, PubDate.Value, ModInfo.ModuleID, Element.EntityID.ToString(), Content));
                }
            }

            return SearchItems;
        }

        #endregion

        #region Zone / VDB Handling

        /// <summary>
        /// Returns the ZoneID from PortalSettings
        /// </summary>
        /// <param name="PortalID"></param>
        /// <returns></returns>
        public int? GetZoneID(int PortalID)
        {
            var ZoneSettingKey = SexyContent.PortalSettingsPrefix + "ZoneID";
            var c = PortalController.GetPortalSettingsDictionary(PortalID);

            if (c.ContainsKey(ZoneSettingKey))
                return int.Parse(c[ZoneSettingKey]);
            return null;
        }

        /// <summary>
        /// Sets the ZoneID from PortalSettings
        /// </summary>
        /// <param name="ZoneID"></param>
        /// <param name="PortalID"></param>
        public void SetZoneID(int? ZoneID, int PortalID)
        {
            if (ZoneID.HasValue)
                PortalController.UpdatePortalSetting(PortalID, SexyContent.PortalSettingsPrefix + "ZoneID", ZoneID.Value.ToString());
            else
                PortalController.DeletePortalSetting(PortalID, SexyContent.PortalSettingsPrefix + "ZoneID");
        }

        #endregion

        #region Apps

        /// <summary>
        /// Returns the default app id
        /// </summary>
        /// <returns></returns>
        public int GetDefaultAppID()
        {
            var ZoneID = GetZoneID(PortalSettings.Current.PortalId);
            return ContentContext.Apps.Where(a => a.Name == EavContext.DefaultAppName && a.ZoneID == ZoneID).Select(a => a.AppID).First();
        }

        #endregion

        #region Culture Handling

        [Obsolete("Don't use this anymore, use 'GetCurrentLanguageName' (work with strings)")]
        public int? GetCurrentLanguageID(bool UseDefaultLanguageIfNotFound = false)
        {
            var LanguageID = ContentContext.GetLanguageId(System.Threading.Thread.CurrentThread.CurrentCulture.Name);
            if (!LanguageID.HasValue && UseDefaultLanguageIfNotFound)
                LanguageID = ContentContext.GetLanguageId(PortalSettings.Current.DefaultLanguage);
            return LanguageID;
        }

        public string GetCurrentLanguageName()
        {
            return System.Threading.Thread.CurrentThread.CurrentCulture.Name;
        }

        public void SetCultureState(string CultureCode, bool Active, int PortalID)
        {
            var EAVLanguage = ContentContext.GetLanguages().Where(l => l.ExternalKey == CultureCode).FirstOrDefault();
            // If the language exists in EAV, set the active state, else add it
            if (EAVLanguage != null)
                ContentContext.UpdateDimension(EAVLanguage.DimensionID, Active);
            else
            {
                var CultureText = LocaleController.Instance.GetLocale(CultureCode).Text;
                ContentContext.AddLanguage(CultureText, CultureCode);
            }
        }


        /// <summary>
        /// Returns all DNN Cultures with active / inactive state
        /// </summary>
        public List<CulturesWithActiveState> GetCulturesWithActiveState(int PortalID, int ZoneID)
        {
            //var DefaultLanguageID = ContentContext.GetLanguageId();
            var AvailableEAVLanguages = ContentContext.GetLanguages();
            var DefaultLanguageCode = new PortalSettings(PortalID).DefaultLanguage;
            var DefaultLanguage = AvailableEAVLanguages.Where(p => p.ExternalKey == DefaultLanguageCode).FirstOrDefault();
            var DefaultLanguageIsActive = DefaultLanguage != null && DefaultLanguage.Active;

            return (from c in LocaleController.Instance.GetLocales(PortalID)
                    select new CulturesWithActiveState()
                    {
                        Code = c.Value.Code,
                        Text = c.Value.Text,
                        Active = AvailableEAVLanguages.Any(a => a.Active && a.ExternalKey == c.Value.Code && a.ZoneID == ZoneID),
                        // Allow State Change only if
                        // 1. This is the default language and default language is not active or
                        // 2. This is NOT the default language and default language is active
                        AllowStateChange = (c.Value.Code == DefaultLanguageCode && !DefaultLanguageIsActive) || (DefaultLanguageIsActive && c.Value.Code != DefaultLanguageCode)
                    }).OrderByDescending(c => c.Code == DefaultLanguageCode).ThenBy(c => c.Code).ToList();

        }

        public class CulturesWithActiveState
        {
            public string Code { get; set; }
            public string Text { get; set; }
            public bool Active { get; set; }
            public bool AllowStateChange { get; set; }
        }

        #endregion

        #region Helper Methods

        private string StripHtmlAndHtmlDecode(string Text)
        {
            return HttpUtility.HtmlDecode(Regex.Replace(Text, "<.*?>", string.Empty));
        }

        public void AddDNNVersionToBodyClass(Control Parent)
        {
            // Add DNN Version to body as CSS Class
            string CssClass = "dnn-" + Assembly.GetAssembly(typeof(DotNetNuke.Common.Globals)).GetName().Version.Major;
            HtmlGenericControl body = (HtmlGenericControl)Parent.Page.FindControl("ctl00$body");
            if(body.Attributes["class"] != null)
                body.Attributes["class"] += CssClass;
            else
                body.Attributes["class"] = CssClass;
        }

        /// <summary>
        /// ToDo: Temp
        /// </summary>
        /// <param name="elements"></param>
        /// <param name="dimensionID"></param>
        /// <returns></returns>
        public string GetJsonFromElements(List<Element> elements, string language)
        {
            var x = new
            {
                Default = new
                {
                    List = (from c in elements
                            where c.EntityID.HasValue
                            select new
                            {
                                Content = GetDictionaryFromEntity(((DynamicEntity)c.Content).Entity, language)
                            }).ToList()
                }
            };

            return x.ToJson();
        }

        private Dictionary<string, object> GetDictionaryFromEntity(IEntity entity, string language)
        {
            var dictionary = ((from d in entity.Attributes select d).ToDictionary(k => k.Value.Name, v => v.Value[language]));
            dictionary.Add("EntityId", entity.EntityId);
            return dictionary;
        }

        #endregion
    }
}